#include <avr/io.h>

#include "snake.h"

snake_t snake;

// Richtungen
// NORT ^ SUED = 2
// WEST ^ OST  = 2
// ++ dreht nach links  (math. pos.)
// -- dreht nach rechts (math. neg.)
#define SN_N 3 // Nord
#define SN_S 1 // Sued
#define SN_W 0 // West
#define SN_E 2 // Ost

#define WALL_SKIP 2

#include "atan2tab.h"

// (A-B)/4 berechnen und so runden, dass das Vorzeichen gleich bleibt.
// ./4 damit die Differenz in 8 Bit (signed) passt
int8_t subr (uint8_t a, uint8_t b)
{
    return (a < b)
        ? (a-b) >> 2
        : -((b-a) >> 2);
}

void snake_umlauf (snake_t * s, uint8_t dir, uint8_t len)
{
    // P0: Beginn des Mauerelements
    uint8_t x0 = s->draw.x;
    uint8_t y0 = s->draw.y;

    // Laenge des Mauerelements
    len <<= WALL_SKIP;

    // P1: Ende des Mauerelements
    uint8_t x1 = x0;
    uint8_t y1 = y0;
    if (dir < 2)
        len = -len;
    if (dir & 1)    y1 += len;
    else            x1 += len;

    // Richtungsvektoren vom Punkt, fuer den der Umlauf
    // zu bestimmen ist (s->food) zu den Endpunkten des
    // Mauersegments
    // P2 = P0 - Food
    // P3 = P1 - Food
    int8_t x2 = subr (x0, s->food.x);
    int8_t y2 = subr (y0, s->food.y);
    int8_t x3 = subr (x1, s->food.x);
    int8_t y3 = subr (y1, s->food.y);

    // Element so hindrehen, da P3 auf (K,0) mit K >= 0
    // abgebildet wird. Matrix ist M=((x3,y3),(y3,-x3))
    // P4 = M*P2
    // M*P3 zeigt in positive x-Richtung
    int16_t x4 = x3*x2 + y3*y2;
    int16_t y4 = y3*x2 - x3*y2;

    // Auf 8 Bits (signed) zusammenschieben
    // P4 >>= ?
    while (1)
    {
        if (x4 < 128 && x4 > -128)
            if (y4 < 128 && y4 > -128)
                break;
        x4 >>= 1;
        y4 >>= 1;
    }

    // Auf 4 Bits (signed) zusammenschieben
    // P5 = P4 >> ?
    int8_t x5 = x4;
    int8_t y5 = y4;

    while (1)
    {
        if (x5 < 8 && x5 > -8)
            if (y5 < 8 && y5 > -8)
                break;
        x5 >>= 1;
        y5 >>= 1;
    }

    // Winkel von P5 gegen die positive X-Achse bestimmen
    // Das ist der Winkel, unter dem (P0,P1) von Food aus
    // erscheint.

    // Index in atan2-Tabelle errechnen
    uint8_t x6 = x5+7;
    uint8_t y6 = y5+7;
    uint8_t index = (uint8_t) (x6 + 15*y6);

    // Winkel aus Tabelle lesen (Vollwinkel ~ 100)
    int8_t phi = pgm_read_byte (& atan2tab[index]);

    s->wall_umlauf += phi;
}
